from warnings import warn

import numpy as np
import xarray as xr
from sklearn.base import BaseEstimator
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.preprocessing import StandardScaler

from ..samples import SampleSpace


class GPEmulator(BaseEstimator):
    def __init__(self, n_targets=2, random_state=None, kernel=None):
        self.random_state = random_state
        self.kernel = kernel
        self.n_targets = n_targets
        self.gps = None
        self.scaler = StandardScaler()
        self.__mean_y = None
        self.__std_y = None

    def fit(self, X, y, y_err):
        if X.shape[-1] != self.n_targets:
            raise ValueError(
                "Last dimension of x do not agree with the specified n_targets"
            )
        if len(y.shape) == 1:
            y = y[:, None]
        if len(y_err.shape) == 1:
            y_err = y_err[:, None]
        if len(y) != 1:
            self.__mean_y = np.mean(y, axis=0)
            self.__std_y = np.std(y, axis=0)
            y = (y - self.__mean_y) / self.__std_y
            y_err = y_err / self.__std_y
        else:
            # We are not standardizing here.
            self.__mean_y = y[0]
            self.__std_y = np.ones(y[0].shape)

        self.scaler.fit(X)
        X = self.scaler.transform(X)

        self.gps = [
            GaussianProcessRegressor(
                alpha=y_err[:, i] ** 2,
                kernel=self.kernel,
                random_state=self.random_state,
            )
            for i in range(self.n_targets)
        ]
        for i, gp in enumerate(self.gps):
            gp.fit(X, y[:, i])

    def predict(self, X, return_std=False):
        X = self.scaler.transform(X)
        values = np.array([gp.predict(X, return_std=return_std) for gp in self.gps])
        if return_std:
            return np.transpose(
                (values[:, 0, :] * self.__std_y[:, None]) + self.__mean_y[:, None]
            ), np.transpose(values[:, 1, :] * self.__std_y[:, None])
        return np.transpose(values * self.__std_y[:, None] + self.__mean_y[:, None])

    def predict_over_space(self, space: SampleSpace, return_std=False, resolution=None):
        """
        Run prediction over sample space
        """
        space_up = space.to_xarray(resolution=resolution)
        valid_points = np.array(np.nonzero(space_up.values))
        X = np.zeros(valid_points.shape)
        for i, dim in enumerate(space_up.dims):
            X[i, :] = space_up[dim][valid_points[i, :]]
        X = np.transpose(X)

        # TODO: This ca
        # n almost certainly be implemented better with less code duplication
        if return_std:
            pred, pred_std = self.predict(X, return_std=True)
            predictions = np.empty((self.n_targets, *space_up.shape)) * np.nan
            predictions_std = np.empty((self.n_targets, *space_up.shape)) * np.nan
            for pred_point, pred_std, valid_point in zip(
                pred, pred_std, valid_points.T
            ):
                predictions[(slice(None), *valid_point)] = pred_point
                predictions_std[(slice(None), *valid_point)] = pred_std
            return xr.DataArray(
                predictions,
                dims=("n_features", *space_up.dims),
                coords={"n_features": np.arange(self.n_targets), **space_up.coords},
            ), xr.DataArray(
                predictions_std,
                dims=("n_features", *space_up.dims),
                coords={"n_features": np.arange(self.n_targets), **space_up.coords},
            )

        else:
            pred = self.predict(X, return_std=False)
            predictions = np.empty((self.n_targets, *space_up.shape)) * np.nan
            for pred_point, valid_point in zip(pred, valid_points.T):
                predictions[(slice(None), *valid_point)] = pred_point
            return xr.DataArray(
                predictions,
                dims=("n_features", *space_up.dims),
                coords={"n_features": np.arange(self.n_targets), **space_up.coords},
            )



import numpy as np
import pytest
import xarray

from history_matching.samples import SampleSpace


@pytest.mark.parametrize(
    "bounds_dict,", [{"0": (1, 2)}, {"0": (1, 2), "1": (10, 20), "2": (-10, 10)}]
)
def test_from_bounds_dict(bounds_dict):
    space = SampleSpace.from_bounds_dict(bounds_dict)
    assert isinstance(space, SampleSpace)
    assert space.ndims == len(bounds_dict)
    assert space.bounds_dict == bounds_dict


@pytest.mark.parametrize(
    "bounds,",
    [
        [
            (1, 2),
        ],
        [(1, 2), (2, 3)],
        [(1, 2), (2, 3), (-10, 10)],
        [(1, 2), (2, 3), (-10, 10), (-5, 10423)],
    ],
)
def test_from_bounds(bounds):
    space = SampleSpace.from_bounds(bounds)
    assert isinstance(space, SampleSpace)
    assert space.ndims == len(bounds)
    for i, b in enumerate(bounds):
        assert space.bounds_dict[str(i)] == b


def test_wrong_bound_order():
    with pytest.raises(ValueError) as e:
        space = SampleSpace.from_bounds([(1, 1)])


def test_incorrect_bounds():
    with pytest.raises(ValueError) as e:
        space = SampleSpace.from_bounds([(1, 2, 3)])


@pytest.mark.parametrize(
    "bounds,",
    [
        [
            (1, 2),
        ],
        [(1, 2), (2, 3)],
        [(1, 2), (2, 3), (-10, 10)],
        [(1, 2), (2, 3), (-10, 10), (-5, 10423)],
    ],
)
def test_lhs_space(bounds):
    space = SampleSpace.from_bounds(bounds)
    sample = space.lhs_sample(5, "c")
    assert sample.shape == (5, space.ndims)
    for dim in range(space.ndims):
        assert (sample[:, dim] < bounds[dim][1]).all()
        assert (sample[:, dim] >= bounds[dim][0]).all()


@pytest.mark.parametrize(
    "bounds,",
    [
        [
            (1, 2),
        ],
        [(1, 2), (2, 3)],
        [(1, 2), (2, 3), (-10, 10)],
        [(1, 2), (2, 3), (-10, 10), (-5, 10423)],
    ],
)
def test_uniform_space(bounds):
    space = SampleSpace.from_bounds(bounds)
    sample = space.uniform(5)
    assert sample.shape == (5, space.ndims)
    for dim in range(space.ndims):
        assert (sample[:, dim] < bounds[dim][1]).all()
        assert (sample[:, dim] >= bounds[dim][0]).all()


@pytest.mark.parametrize(
    "dim,",
    [1, 2, 3, 4, 5],
)
def test_from_xarray(dim):
    data = (
        np.sum(np.meshgrid(*[np.linspace(0, 1, 10) for _ in range(dim)]), axis=0) / dim
        > 0.5
    )
    ds = xarray.DataArray(
        data=data, coords={str(d): np.linspace(0, 1, 10) for d in range(dim)}
    )
    space = SampleSpace.from_xarray(ds)
    assert isinstance(space, SampleSpace)
    samples = space.uniform(100)
    assert (samples.sum(axis=-1) > 0.5).all()  # Check samples make sense



import numpy as np
import xarray as xr
from sklearn.base import BaseEstimator
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import RBF, ConstantKernel, WhiteKernel
from sklearn.preprocessing import MinMaxScaler

from ..samples import SampleSpace


class Emulator(BaseEstimator):
    def __init__(
        self, n_targets=2, random_state=None, kernel=None, n_restarts_optimizer=0
    ) -> None:
        self.random_state = random_state

        self.n_targets = n_targets
        if kernel is None:
            self.kernel = (
                ConstantKernel() * RBF(length_scale_bounds=(0.05, 1)) + WhiteKernel()
            )
        else:
            self.kernel = kernel
        self.scaler_x = MinMaxScaler()
        self.n_restarts_optimizer = n_restarts_optimizer
        self.__gps = [
            GaussianProcessRegressor(
                normalize_y=True,
                random_state=self.random_state,
                kernel=self.kernel,
                n_restarts_optimizer=self.n_restarts_optimizer,
            )
            for _ in range(self.n_targets)
        ]
        super().__init__()

    def fit(self, X, y, y_err=None):
        X = self.scaler_x.fit_transform(X)
        if y_err is not None:
            y_err = y_err / (y.std(axis=0) if y.std() != 0 else 1.0)
            self.__gps = [
                GaussianProcessRegressor(
                    alpha=y_err[:, i] ** 2,
                    normalize_y=True,
                    random_state=self.random_state,
                    kernel=self.kernel,
                    n_restarts_optimizer=self.n_restarts_optimizer,
                )
                for i in range(self.n_targets)
            ]
        for i, gp in enumerate(self.__gps):
            gp.fit(X, y[:, i])

    def predict(self, X, return_std=False):
        X = self.scaler_x.transform(X)
        values = np.array([gp.predict(X, return_std=return_std) for gp in self.__gps])
        if return_std:
            return np.transpose(values[:, 0]), np.transpose(values[:, 1])
        return np.transpose(values)

    def predict_over_space(self, space: SampleSpace, return_std=False, resolution=None):
        space_xr = space.to_xarray(resolution=resolution)
        valid_points = np.array(np.nonzero(space_xr.values))
        X = np.zeros(valid_points.shape)
        for i, dim in enumerate(space_xr.dims):
            X[i, :] = space_xr[dim][valid_points[i, :]]
        X = np.transpose(X)
        dims = ("n_features", *space_xr.dims)
        coords = {"n_features": np.arange(self.n_targets), **space_xr.coords}
        predictions = np.empty((self.n_targets, *space_xr.shape)) * np.nan
        if return_std:
            predictions_std = np.empty((self.n_targets, *space_xr.shape)) * np.nan
            preds_flt, preds_std_flt = self.predict(X, return_std=True)
            for pred, pred_std, valid_point in zip(
                preds_flt, preds_std_flt, np.transpose(valid_points)
            ):
                predictions[(slice(None), *valid_point)] = pred
                predictions_std[(slice(None), *valid_point)] = pred_std

            return xr.DataArray(predictions, dims=dims, coords=coords), xr.DataArray(
                predictions_std, dims=dims, coords=coords
            )
        else:
            preds_flt = self.predict(X, return_std=False)
            for pred, valid_point in zip(preds_flt, np.transpose(valid_points)):
                predictions[(slice(None), *valid_point)] = pred
            return xr.DataArray(predictions, dims=dims, coords=coords)
